import time
t0 = time.time()

import logging,inspect
log_and_raise_error = lambda logger, message: log(logger, message, lvl="ERROR", exception=RuntimeError, rollback=2)
log_warning = lambda logger, message: log(logger, message, lvl="WARNING", exception=None, rollback=2)
log_info = lambda logger, message: log(logger, message, lvl="INFO", exception=None, rollback=2)
log_debug = lambda logger, message: log(logger, message, lvl="DEBUG", exception=None, rollback=2)


def log(logger, message, lvl, exception=None, rollback=1):
    logcalls = {"ERROR": logger.error,
                "WARNING": logger.warning,
                "DEBUG": logger.debug,
                "INFO": logger.info}
    if lvl not in logcalls:
        print "%s is an invalid logger level." % lvl
        sys.exit(1)
    logcall = logcalls[lvl]
    # This should maybe go into a handler
    if (logger.getEffectiveLevel() >= logging.INFO) or rollback is None:
        # Short output
        msg = "%s" % message
    else:
        # Detailed output only in debug mode
        func = inspect.currentframe()
        for r in range(rollback):
            # Rolling back in the stack, otherwise it would be this function
            func = func.f_back
        code = func.f_code
        #msg = "%s\n\t=> in \'%s\' function \'%s\' [%s:%i]" % (message,
        #                                                      func.f_globals["__name__"],
        #                                                      code.co_name, 
        #                                                      code.co_filename, 
        #                                                      code.co_firstlineno)
        msg = message
        
    logcall("\t%i sec\t%s" % (time.time()-t0, msg))
    if exception is not None:
        raise exception(message)



import numpy, os, time
import h5py

from log import log_and_raise_error, log_warning, log_info, log_debug

from h5writer import AbstractH5Writer,logger

class H5Writer(AbstractH5Writer):
    """
    HDF5 writer class for single-process writing.
    """
    def __init__(self, filename, chunksize=100, compression=None):
        AbstractH5Writer.__init__(self, filename, chunksize=chunksize, compression=compression)
        if os.path.exists(self._filename):
            log_warning(logger, self._log_prefix + "File %s exists and is being overwritten" % (self._filename))
        self._f = h5py.File(self._filename, "w")
        
    def write_slice(self, data_dict):
        """
        Call this function for writing all data in data_dict as a stack of slices (first dimension = stack dimension).
        Dictionaries within data_dict are represented as HDF5 groups. The slice index is either the next one.
        """
        if not self._initialised:
            # Initialise of tree (groups and datasets)
            self._initialise_tree(data_dict)
            self._initialised = True        
        # Iterate index
        if self._i is None:
            self._i = 0
        else:
            self._i += 1
        # Expand stacks if needed
        if self._i >= (self._stack_length-1):
            self._resize_stacks(self._stack_length * 2)
        # Write data
        self._write_group(data_dict)
        # Update of maximum index
        self._i_max = self._i if self._i > self._i_max else self._i_max

    def write_solo(self, data_dict):
        """
        Call this function for writing datasets that have no stack dimension (i.e. no slices).
        """
        self._write_solo_group(data_dict)

    def _write_solo_group(self, data_dict, group_prefix="/"):
        if group_prefix != "/" and group_prefix not in self._f:
            self._f.create_group(group_prefix)
        keys = data_dict.keys()
        keys.sort()
        for k in keys:
            name = group_prefix + k
            if isinstance(data_dict[k], dict):
                self._write_solo_group(data_dict[k], group_prefix=name+"/")
            else:
                if name in self._f:
                    log_warning(logger, "Dataset %s already exists! Overwriting with new data." % name)
                self._f[name] = data_dict[k]
                
    def close(self):
        """
        Close file.
        """
        self._resize_stacks(self._i_max + 1)
        self._f.close()
        log_info(logger, self._log_prefix + "HDF5 writer instance for file %s closed." % (self._filename))
        



import numpy, os, time
import h5py

import logging
logger = logging.getLogger(__name__)

from log import log_and_raise_error, log_warning, log_info, log_debug

CHUNKSIZE_MIN_IN_BYTES = 16000000

class AbstractH5Writer:
    def __init__(self, filename, chunksize, compression):
        self._filename = os.path.expandvars(filename)        
        self._chunksize = chunksize
        self._stack_length = chunksize
        self._i = None
        self._i_max = -1
        self._create_dataset_kwargs = {}
        self._log_prefix = ""
        if compression is not None:
            self._create_dataset_kwargs["compression"] = compression
        self._initialised = False
        
    def _initialise_tree(self, D, group_prefix="/"):
        keys = D.keys()
        keys.sort()
        for k in keys:
            if isinstance(D[k],dict):
                group_prefix_new = group_prefix + k + "/"
                log_debug(logger, self._log_prefix + "Creating group %s" % (group_prefix_new))
                if group_prefix_new != "/" and group_prefix_new not in self._f:
                    self._f.create_group(group_prefix_new)
                self._initialise_tree(D[k], group_prefix=group_prefix_new)
            else:
                name = group_prefix + k
                log_debug(logger, self._log_prefix + "Creating dataset %s" % (name))
                if name not in self._f:
                    data = D[k]
                    self._create_dataset(data, name)
                    
    def _write_group(self, D, group_prefix="/"):
        keys = D.keys()
        keys.sort()
        for k in keys:
            if isinstance(D[k],dict):
                group_prefix_new = group_prefix + k + "/"
                self._write_group(D[k], group_prefix_new)
            else:
                name = group_prefix + k
                data = D[k]
                log_debug(logger, self._log_prefix + "Write to dataset %s at stack position %i" % (name, self._i))
                if name not in self._f:
                    log_and_raise_error(logger, self._log_prefix + "Write to dataset %s at stack position %i failed because it does not exist. Note that all datasets are initialised from the data of the first write call." % (name, self._i))
                if numpy.isscalar(data):
                    self._f[name][self._i] = data
                else:
                    self._f[name][self._i,:] = data[:]
                
    def _create_dataset(self, data, name):
        data = numpy.asarray(data)
        try:
            h5py.h5t.py_create(data.dtype, logical=1)
        except TypeError:
            log_warning(logger, self._log_prefix + "Could not save dataset %s. Conversion to numpy array failed" % (name))
            return 1
        maxshape = tuple([None]+list(data.shape))
        shape = tuple([self._chunksize]+list(data.shape))
        dtype = data.dtype
        if dtype.type is numpy.string_:
            dtype = h5py.new_vlen(str)
        nbytes_chunk = numpy.prod(shape) * dtype.itemsize
        if nbytes_chunk > CHUNKSIZE_MIN_IN_BYTES:
            chunksize = self._chunksize
            #log_debug(logger, self._log_prefix + "Do not increase chunksize (%i) for dataset %s (%i bytes for single data frame)" % (self._chunksize, name, nbytes_chunk))
        else:
            chunksize = int(numpy.ceil(float(CHUNKSIZE_MIN_IN_BYTES) / float(data.nbytes)))
            log_debug(logger, self._log_prefix + "Increase chunksize from %i to %i for dataset %s (only %i bytes for single data frame)" % (self._chunksize, chunksize, name, nbytes_chunk))
        chunks = tuple([chunksize]+list(data.shape))
        ndim = data.ndim
        axes = "experiment_identifier"
        if ndim == 1: axes = axes + ":x"
        elif ndim == 2: axes = axes + ":y:x"
        elif ndim == 3: axes = axes + ":z:y:x"
        log_debug(logger, self._log_prefix + "Create dataset %s [shape=%s, chunks=%s, dtype=%s]" % (name, str(shape), str(chunks), str(dtype)))
        self._f.create_dataset(name, shape, maxshape=maxshape, dtype=dtype, chunks=chunks)
        self._f[name].attrs.modify("axes",[axes])
        return 0

    def _is_stack(self, name):
        a = self._f[name].attrs
        return ("axes" in a.keys() and a["axes"][0].startswith('experiment_identifier'))
    
    def _resize_stacks(self, stack_length, group_prefix="/"):
        if group_prefix == "/":
            log_info(logger, self._log_prefix + "Resize datasets to new length: %i" % stack_length)
        if stack_length == 0:
            log_warning(logger, self._log_prefix + "Cannot resize stacks to length 0. Skip resize stacks.")
            return
        keys = self._f[group_prefix].keys()
        keys.sort()
        for k in keys:
            name = group_prefix + k
            if isinstance(self._f[name], h5py.Dataset):
                if self._is_stack(name):
                    self._resize_stack(stack_length, name)
            else:
                self._resize_stacks(stack_length, name + "/")
        self._stack_length = stack_length
            
    def _resize_stack(self, stack_length, name):
        new_shape = list(self._f[name].shape)
        new_shape[0] = stack_length
        new_shape = tuple(new_shape)
        log_debug(logger, self._log_prefix + "Resize dataset %s [old shape: %s, new shape: %s]" % (name, str(self._f[name].shape), str(new_shape)))
        t0 = time.time()
        self._f[name].resize(new_shape)
        t1 = time.time()
        log_debug(logger, self._log_prefix + "Resizing time: %f sec (HDF5)" % (t1-t0))



import numpy, os, time
import h5py

try:
    from mpi4py import MPI
except:
    log_warning(logger, "Cannot import mpi4py!")    
    MPI = None
    
from log import log_and_raise_error, log_warning, log_info, log_debug

from h5writer import AbstractH5Writer,logger

class H5WriterMPISW(AbstractH5Writer):
    """
    HDF5 writer class for MPI with a master process (rank=0) reserved for writing and inter-process-communication. Slave processes (rank>0) can call write methods and data is sent to the master process for serial writing.
    """
    def __init__(self, filename, comm, chunksize=100, compression=None):
        if MPI is None:
            log_and_raise_error(logger, "Could not import MPI and hence cannot initialise H5WriterMPI instance.")
            return
        self.comm = comm
        if not isinstance(self.comm, MPI.Comm):
            log_and_raise_error(logger, "Cannot initialise H5WriterMPI instance. \'%s\' is not an mpi4py.MPI.Comm instance." % str(self.comm))
        AbstractH5Writer.__init__(self, filename, chunksize=chunksize, compression=compression)
        self._closed = False
        if self._is_master():
            if os.path.exists(self._filename):
                log_warning(logger, self._log_prefix + "File %s exists and is being overwritten" % (self._filename))
            self._f = h5py.File(self._filename, "w")
            self._master_loop()
            
    def _is_in_communicator(self):
        try:
            return self.comm.rank != MPI.UNDEFINED
        except MPI.Exception:
            return False
            
    def _is_master(self):
        return (self._is_in_communicator() and self.comm.rank == 0)

    def _master_loop(self):
        log_debug(logger, "Master loop started")
        t_start = time.time()
        slices = numpy.zeros(self.comm.size, 'i')
        closed = numpy.zeros(self.comm.size, 'i')
        while True:

            status = MPI.Status()
            l = self.comm.recv(source=MPI.ANY_SOURCE, tag=0, status=status)
            source = status.Get_source()

            if source == 0:
                logger.warning('Received write package from master process! Skipping writing.')
                print l
                continue

            if l == "close":
                closed[source] = 1
                if closed.sum() == self.comm.size-1:
                    break

            else:
                self._transfer_numpy_arrays(l, source=source)
                if "write_slice" in l:
                    log_debug(logger, "Write slice to file")
                    t0 = time.time()
                    self._write_slice_master(l["write_slice"], i=slices[source]*(self.comm.size-1)+source-1)
                    slices[source] += 1
                    t1 = time.time()
                    t_log = t1-t0
                    log_info(logger, "Writing rate %.1f Hz; slice %i; logging %.2f sec" % (slices.sum()/(time.time()-t_start),slices.sum(),t_log))
                if "write_solo" in l:
                    log_debug(logger, "Write solo to file")
                    self._write_solo_master(l["write_solo"])

        log_debug(logger, "Master writer is closing.")
        self._resize_stacks(self._i_max + 1)
        self._f.close()
        log_debug(logger, "File %s closed." % self._filename)
                        
    def _send_for_writing(self, data_dict):
        """
        Sending data dictionaries to master process without pickeling numpy arrays.
        """
        skeleton = _make_skeleton(data_dict)
        self.comm.send(skeleton, dest=0, tag=0)
        self._transfer_numpy_arrays(skeleton=skeleton, data_dict=data_dict)

    def _transfer_numpy_arrays(self, skeleton, data_dict=None, source=None):
        mode = 'master' if data_dict is None else 'slave' 
        keys = skeleton.keys()
        keys.sort()
        for k in keys:
            if isinstance(skeleton[k], dict):
                self._transfer_numpy_arrays(skeleton[k], None if mode == 'master' else data_dict[k], source=source)
            elif isinstance(skeleton[k], _ArrayDescriptor):
                if mode == 'master':
                    skeleton[k] = numpy.empty(shape=skeleton[k].shape, dtype=skeleton[k].dtype)
                    self.comm.Recv(skeleton[k], source=source, tag=0)
                else:
                    self.comm.Send(data_dict[k], dest=0, tag=0)

    def write_slice(self, data_dict):
        """
        Call this function for writing all data in data_dict as a stack of slices (first dimension = stack dimension).
        Dictionaries within data_dict are represented as HDF5 groups. The slice index is the next one.
        """
        self._send_for_writing({"write_slice": data_dict})
        #self.comm.send({"write_slice": data_dict}, dest=0, tag=0)
        
    def _write_slice_master(self, data_dict, i):
        if not self._initialised:
            # Initialise of tree (groups and datasets)
            self._initialise_tree(data_dict)
            self._initialised = True        
        self._i = i
        # Expand stacks if needed
        while self._i >= (self._stack_length-1):
            self._resize_stacks(self._stack_length * 2)
        # Write data
        self._write_group(data_dict)
        # Update of maximum index
        self._i_max = self._i if self._i > self._i_max else self._i_max

    def write_solo(self, data_dict):
        """
        Call this function for writing datasets that have no stack dimension (i.e. no slices).
        """
        self._send_for_writing({"write_solo": data_dict})
        #self.comm.send({"write_solo": data_dict}, dest=0, tag=0)
        
    def _write_solo_master(self, data_dict):
        self._write_solo_group(data_dict)

    def _write_solo_group(self, data_dict, group_prefix="/"):
        if group_prefix != "/" and group_prefix not in self._f:
            self._f.create_group(group_prefix)
        keys = data_dict.keys()
        keys.sort()
        for k in keys:
            name = group_prefix + k
            if isinstance(data_dict[k], dict):
                self._write_solo_group(data_dict[k], group_prefix=name+"/")
            else:
                if name in self._f:
                    log_warning(logger, "Dataset %s already exists! Overwriting with new data." % name)
                self._f[name] = data_dict[k]
                
    def close(self):
        """
        Close file.
        """
        if self._closed:
            log_debug("Instance already closed.")
        if not self._is_master():
            self.comm.send("close", dest=0, tag=0)
        else:
            log_info(logger, self._log_prefix + "HDF5 writer master process for file %s closed." % (self._filename))
        self._closed = True
        
def _make_skeleton(data_dict):
    skeleton = {}
    keys = data_dict.keys()
    keys.sort()
    for k in keys:
        v = data_dict[k]
        if isinstance(v, dict):
            skeleton[k] = _make_skeleton(v)
        else:
            if isinstance(v, numpy.ndarray):
                skeleton[k] = _ArrayDescriptor(v)
            else:
                skeleton[k] = v
    return skeleton

class _ArrayDescriptor:
    def __init__(self, ndarray):
        self.dtype = ndarray.dtype
        self.shape = ndarray.shape



#!/usr/bin/env python

from setuptools import setup

setup(name='h5writer',
      version='0.4.9',
      description='Writing HDF5 files with openMPI.',
      author='Max F. Hantke, Benedikt Daurer',
      author_email='maxhantke@gmail.com',
      url='https://github.com/mhantke/h5writer',
      install_requires=['numpy', 'h5py', 'mpi4py>=2.0.0'],
      packages = ['h5writer'],
)




import sys
import numpy, os, time
import h5py

from log import log_and_raise_error, log_warning, log_info, log_debug

try:
    from mpi4py import MPI
except:
    log_warning(logger, "Cannot import mpi4py!")

MPI_TAG_INIT   = 1
MPI_TAG_EXPAND = 2
MPI_TAG_READY  = 3
MPI_TAG_CLOSE  = 4

from h5writer import AbstractH5Writer,logger

class H5WriterMPIPW(AbstractH5Writer):
    """
    HDF5 writer class for MPI and parallel writing.
    """
    def __init__(self, filename, comm, chunksize=100, compression=None):
        # Initialisation of base class
        AbstractH5Writer.__init__(self, filename, chunksize=chunksize, compression=compression)
        # MPI communicator
        self.comm = comm
        # This "if" avoids that processes that are not in the communicator (like the master process of hummingbird) interact with the file and block
        if not self._is_in_communicator():
            log_warning(logger, "This process cannot write.")
            return
        # Logging
        self._log_prefix = "(%03i) " %  self.comm.rank
        # Index
        self._i = -self.comm.size + self.comm.rank
        self._i_max = -1
        # Expansion flag
        self._expand_flag = False
        # Chache
        self._solocache = {}
        # Status
        self._ready = False
        # Open file
        if os.path.exists(self._filename):
            log_warning(logger, self._log_prefix + "File %s exists and is being overwritten" % (self._filename))
        self.comm.Barrier()
        self._f = h5py.File(self._filename, "w", driver='mpio', comm=self.comm)
        self._f.atomic = True
        self.comm.Barrier()
        log_debug(logger, self._log_prefix + "File successfully opened for parallel writing.")
        
    def write_slice(self, data_dict):
        """
        Call this function for writing all data in data_dict as a stack of slices (first dimension = stack dimension).
        Dictionaries within data_dict are represented as HDF5 groups. The slice index is either the next one.
        """

        # Initialise of tree (groups and datasets)
        if not self._initialised:
            self.comm.Barrier()
            self._initialise_tree(data_dict)
            if self.comm.size > self._stack_length:
                self.comm.Barrier()
                self._expand_stacks_mpi(i_max=self.comm.size-1)  
            self._initialised = True
            self.comm.Barrier()
            log_debug(logger, self._log_prefix + "Tree initialised")
        self._i += self.comm.size
        # Update of maximum index
        if self._i > self._i_max:
            self._i_max = self._i
        # Enter poll for other processes that might need stack expansion?
        self._expand_poll()
        log_debug(logger, self._log_prefix + "Write")
        # Write data
        self._write_group(data_dict)
        log_debug(logger, self._log_prefix + "Written")
        sys.stdout.flush()

    def write_solo(self, data_dict):
        """
        Call this function for writing datasets that have no stack dimension (i.e. no slices).
        """
        if self._is_in_communicator():
            return self._to_solocache(data_dict, target=self._solocache)
        
    def write_solo_mpi_reduce(self, data_dict, op):
        """
        Call this function for writing datasets that have no stack dimension (i.e. no slices).
        Data will be reduced between processes using the given open MPI operator (see for example https://pythonhosted.org/mpi4py/apiref/mpi4py-module.html).
        """
        if self._is_in_communicator():
            return self._to_solocache(data_dict, target=self._solocache, op=op)

    def close(self):
        """
        Close file.
        """
        # This "if" avoids that processes that are not in the communicator (like the master process of hummingbird) interact with the file and block
        if not self._is_in_communicator():
            return

        if not self._initialised:
            log_and_raise_error(logger, "Cannot close uninitialised file. Every worker has to write at least one frame to file. Reduce your number of workers and try again.")
            exit(1)
        self._close_signal()
        log_debug(logger, self._log_prefix + "Enter closing loop")
        while True:
            self._expand_poll()
            self._update_ready()
            if self._ready:
                break
            time.sleep(0.1)
        log_debug(logger, self._log_prefix + "Leave closing loop")

        self.comm.Barrier()
        log_debug(logger, self._log_prefix + "Sync stack length")
        self._sync_i_max()

        self.comm.Barrier()
        log_debug(logger, self._log_prefix + "Closing file %s for parallel writing." % (self._filename))
        self._f.close()
        log_debug(logger, self._log_prefix + "File %s closed for parallel writing." % (self._filename))

        if self._is_master():
            log_debug(logger, self._log_prefix + "Opening file %s for single-process writing." % (self._filename))
            self._f = h5py.File(self._filename, "r+")
            log_debug(logger, self._log_prefix + "Shrink stacks")
            self._resize_stacks(self._i_max + 1)
            
        self._write_solocache_group_to_file(self._solocache)

        if self._is_master():
            log_debug(logger, self._log_prefix + "Closing file %s for single-process writing." % (self._filename))
            self._f.close()
            log_info(logger, self._log_prefix + "HDF5 parallel writer instance for file %s closed." % (self._filename))
    
    def _is_in_communicator(self):
        try:
            out = self.comm.rank != MPI.UNDEFINED
        except MPI.Exception:
            out = False
        return out
            
    def _is_master(self):
        return (self._is_in_communicator() and self.comm.rank == 0)
        
    def _to_solocache(self, data_dict, target, op=None):
        keys = data_dict.keys()
        keys.sort()
        for k in keys:
            if isinstance(data_dict[k], dict):
                if k not in target:
                    target[k] = {}
                self._to_solocache(data_dict[k], target=target[k], op=op)
            else:
                target[k] = (data_dict[k], op)
            
    def _write_solocache_group_to_file(self, data_dict, group_prefix="/"):
        if self._is_master() and group_prefix != "/" and group_prefix not in self._f:
            self._f.create_group(group_prefix)
        keys = data_dict.keys()
        keys.sort()
        for k in keys:
            name = group_prefix + k
            if isinstance(data_dict[k], dict):
                self._write_solocache_group_to_file(data_dict[k], group_prefix=name+"/")
            else:
                (data, op) = data_dict[k]
                if op is not None:
                    if numpy.isscalar(data):
                        sendobj = numpy.array(data)
                    else:
                        sendobj = data
                    recvobj = numpy.empty_like(data)
                    log_debug(logger, self._log_prefix + "Reducing data %s" % (name))
                    self.comm.Reduce(
                        [sendobj, MPI.DOUBLE],
                        [recvobj, MPI.DOUBLE],
                        op = op,
                        root = 0
                    )
                    data = recvobj
                if self._is_master():
                    log_debug(logger, self._log_prefix + "Writing data %s" % (name))
                    self._f[name] = data
        
    def _expand_poll(self):
        #log_debug(logger, self._log_prefix + "Polling for stack expansion")
        expand = False
        if self.comm.Iprobe(source=(self.comm.rank-1) % self.comm.size, tag=MPI_TAG_EXPAND):
            buf1 = numpy.empty(1, dtype='i')
            self.comm.Recv([buf1, MPI.INT], source=(self.comm.rank-1) %  self.comm.size, tag=MPI_TAG_EXPAND)
            buf2 = numpy.array(1, dtype='i')
            self.comm.Send([buf2, MPI.INT], dest=(self.comm.rank+1) %  self.comm.size, tag=MPI_TAG_EXPAND)
            expand= True
        # Sending of expansion signal needed?
        elif self._i >= self._stack_length:
            buf1 = numpy.array(1, dtype='i')
            req_s = self.comm.Isend([buf1, MPI.INT], dest=(self.comm.rank+1) %  self.comm.size, tag=MPI_TAG_EXPAND)
            buf2 = numpy.empty(1, dtype='i')
            req_r = self.comm.Irecv([buf2, MPI.INT], source=(self.comm.rank-1) %  self.comm.size, tag=MPI_TAG_EXPAND)
            while True:
                sent     = req_s.Test()
                received = req_r.Test()
                if sent and received:
                    break
                time.sleep(0.01)
            expand = True
        if expand:
            log_debug(logger, self._log_prefix + "Do stack expansion")
            self.comm.Barrier()
            self._sync_i_max()
            self._expand_stacks_mpi()
        else:
            pass
            #log_debug(logger, self._log_prefix + "No stack expansion")

    def _expand_stacks_mpi(self, i_max=None):
        self._f.close()
        log_debug(logger, self._log_prefix + "File closed for stack expansion")
        self.comm.Barrier()
        if i_max is None:
            i_max = self._i_max
        stack_length_new = self._stack_length
        while i_max >= stack_length_new:
            stack_length_new *= 2
        if self.comm.rank == 0:
            self._f = h5py.File(self._filename, "r+")
            log_debug(logger, self._log_prefix + "Start stack expansion (%i >= %i) - new stack length will be %i" % (i_max, self._stack_length, stack_length_new))
            self._resize_stacks(stack_length_new)
            self._f.close()
        self._stack_length = stack_length_new
        self.comm.Barrier()
        self._f = h5py.File(self._filename, "r+", driver='mpio', comm=self.comm)
        self._f.atomic = True
        self.comm.Barrier()
        log_debug(logger, self._log_prefix + "File reopened after stack expansion")
        
    def _close_signal(self):
        if self.comm.rank == 0:
            self._closing_ranks = [0]
        else:
            buf = numpy.array(self.comm.rank, dtype="i")
            self.comm.Send([buf, MPI.INT], dest=0, tag=MPI_TAG_READY)
            log_debug(logger, self._log_prefix + "Rank %i sent closing signal to master" % (self.comm.rank))
            
    def _update_ready(self):
        if self.comm.rank == 0:
            while self.comm.Iprobe(source=MPI.ANY_SOURCE, tag=MPI_TAG_READY):
                buf = numpy.empty(1, dtype='i')
                self.comm.Recv([buf, MPI.INT], source=MPI.ANY_SOURCE, tag=MPI_TAG_READY)
                rank = buf[0]
                self._closing_ranks.append(rank)
                log_debug(logger, self._log_prefix + "Received closing signal from rank %i (%i/%i)" % (rank,len(self._closing_ranks),self.comm.size))
            if len(self._closing_ranks) == self.comm.size:
                for i in range(1, self.comm.size):
                    send_buf = numpy.array(1, dtype='i')
                    self.comm.Send([send_buf, MPI.INT], dest=i, tag=MPI_TAG_READY)
                    recv_buf = numpy.empty(1, dtype='i')
                    self.comm.Recv([recv_buf, MPI.INT], source=i, tag=MPI_TAG_READY)
                self._ready = True
                log_debug(logger, self._log_prefix + "Master sent out ready signals to slaves")
        else:
            if self.comm.Iprobe(source=0, tag=MPI_TAG_READY):
                recv_buf = numpy.empty(1, dtype='i')
                self.comm.Recv([recv_buf, MPI.INT], source=0, tag=MPI_TAG_READY)
                send_buf = numpy.array(1, dtype='i')
                self.comm.Send([send_buf, MPI.INT], dest=0, tag=MPI_TAG_READY)
                self._ready = True
                log_debug(logger, self._log_prefix + "Slave rank %i received ready signals from master" %  self.comm.rank)
                
    def _sync_i_max(self):
        sendbuf = numpy.array(self._i_max, dtype='i')
        recvbuf = numpy.empty(1, dtype='i')
        log_debug(logger, self._log_prefix + "Entering allreduce with maximum index %i" % (self._i_max))
        self.comm.Allreduce([sendbuf, MPI.INT], [recvbuf, MPI.INT], op=MPI.MAX)
        self._i_max = recvbuf[0]
        log_debug(logger, self._log_prefix + "After reduce: i_max = %i" % self._i_max)

        
    
